(* Alpha Prolog *)
(* Abstract syntax *)
open Types;;
open Printer;;
open Fixity;;

type name_sort = Nstbl.path;;
type data_sort = Nstbl.path;;


type ty = IdTy of Nstbl.path
  | NameTy of name_sort
  | DataTy of data_sort * ty list 
  | UnitTy
  | PairTy of ty * ty
  | AbsTy of ty * ty
  | ListTy of ty
  | VarTy of var
  | UnderscoreTy
  | IntTy
  | CharTy
  | BoolTy
  | PropTy 
  | ArrowTy of ty * ty
;;


type kind = TypeK
  | NameK (* illegal as result kind *)
  | ArrowK of kind * kind (* higher-order kinds illegal *)
;;


type constr = (name*var) list;;


type atomic = Nstbl.path * term list
and term = Atomic of atomic
  | Var of var
  | Name of name
  | Transpose of name * name * term
  | Nil
  | Cons of term * term
  | Unit
  | Pair of term * term
  | Abs of name * term
  | True
  | Fresh of term * term
  | Eq of term * term
  | Is of term * term
  | Implies of term * term
  | Or of term * term
  | And of term * term
  | Forall of var * ty * term
  | Exists of var * ty * term
  | New of var * ty * term 
  | Cut 
  | Guard of term * term * term
  | Not of term
  | Constr of term * constr
  | Underscore
  | IntC of int
  | CharC of char
  | BoolC of bool

;;


type decl = {pos:Pos.pos option; rdecl:rdecl}
and rdecl = KindDecl of sym * kind
  | SymDecl of sym * ty
  | FuncDecl of sym * ty list * ty
  | PredDecl of sym * ty list
  | ClauseDecl of term 
  | Query of term
  | UseDirective of string
  | TraceDirective of int
  | InfixDecl of direction * string * int
  | NamespaceDecl of sym * decl list
  | TypeDefn of sym * var list * ty 
  | BuiltinFuncDecl of sym * ty list * ty 
	* (Isym.term_sym Internal.term list -> Isym.term_sym Internal.term)
  | TypeQuery of term
  | QuitDirective
  | OpenDirective of Nstbl.path
  | HelpDirective of Nstbl.path option
;;


let rec pp_kind k = 
  match k with
    NameK -> text "name_type"
  | TypeK -> text "type"
  | ArrowK(k1,k2) -> pp_kind k1 <+> arrow <+> pp_kind k2
;;




(* TODO: Make associativity less of a hack. *)
let toplevel_prec = 0;;
let forall_prec = 0;;
let exists_prec = 0;;
let new_prec = 0;;
let guard_prec = 0;;
let implies_prec = 1;;
let arrow_prec = 1;;
let constr_prec = 1;;
let or_prec = 2;;
let and_prec = 3;;
let pair_prec = 3;;
let eq_prec = 4;;
let fresh_prec = 4;;
let is_prec = 4;;
let plus_prec = 5;;
let minus_prec = 5;;
let cons_prec = 6;;
let abs_prec = 7;;
let trans_prec = 10;;
let app_prec = 10;;
let not_prec = 10;;


let pp_ty s = 
  let rec pp prec s =
    let paren p d = if p <= prec then Printer.paren(d) else d in
    match s with
      NameTy s -> Nstbl.pp_path s
    | DataTy (ds,ss) -> 
	paren app_prec 
	  (Nstbl.pp_path ds <:> 
	   (hcat (List.map (fun s -> space <:> pp app_prec s) ss)))
    | VarTy v -> text (Var.to_string' v)
    | ListTy s -> bracket(pp toplevel_prec s)
    | AbsTy (_,_) -> paren abs_prec (pp_abss s)
    | PairTy(_,_) -> Printer.paren (pp_pairs s)
    | IdTy s -> Nstbl.pp_path s
    | UnitTy -> text "unit"
    | UnderscoreTy -> underscore
    | IntTy -> text "int"
    | CharTy -> text "char"
    | PropTy -> text "prop"
    | BoolTy -> text "bool"
    | ArrowTy(_,_) -> paren arrow_prec (pp_arrows s)
  and pp_abss s = 
    match s with 
      AbsTy(s1,s2) -> pp abs_prec s1 <:> backslash <:> pp_abss s2
    | s -> pp and_prec s
  and pp_pairs s = 
    match s with 
      PairTy(s1,s2) -> pp and_prec s1 <:> comma <:> pp_pairs s2
    | s -> pp and_prec s
  and pp_arrows s = 
    match s with 
      ArrowTy(s1,s2) -> pp arrow_prec s1 <+> arrow <+> pp_arrows s2
    | s -> pp arrow_prec s
  in pp toplevel_prec s
;;


let rec pp_constr c = 
  match c with
    [] -> emp
  | (a,v)::c -> 
      Var.pp_var (a) 
	<+> hash
	<+> Var.pp_var (v) 
	<:> newline
	<:> pp_constr c

and pp_term term = 
  let pp_clause prec (p,cs,p') = 
    pp_term p <+>
    if not(cs = []) 
    then 
      (slash <+> 
       sep (comma) 
	 (List.map (fun (a,tm) -> 
	   text (Var.to_string' a)  <+> 
	   hash <+> 
	   pp_term tm) cs))
	<+>
      (larrow <+> 
     pp_term p')
    else emp
  in	
  let rec pp_atomic (h,tl) = 
    sep space ((Nstbl.pp_path h)::(List.map (pp app_prec) tl))
  and pp prec term = 
    let paren p d = if p <= prec then Printer.paren(d) else d in
    match term with
      Unit -> text "()"
    | Name a -> Var.pp_var a
    | Var v -> Var.pp_var v
    | Transpose(a,b,(Transpose(_,_,_) as t)) -> 
	Printer.paren (Var.pp_var a <:> tilde <:> Var.pp_var b) 
	  <:> pp toplevel_prec t
     | Transpose(a,b,t) -> 
	Printer.paren (Var.pp_var a <:> tilde <:> Var.pp_var b) 
	  <:> pp trans_prec t
    | Atomic (a,[]) -> (Nstbl.pp_path a)
    | Atomic (a,[t1;t2]) -> (* maybe infix... TODO: fix for overlapping op's *)
	let sym = 
	  match a with
	    Nstbl.Root(s) -> Nstbl.get_base s
	  | Nstbl.Rel(s) -> Nstbl.get_base s
	in
	if List.mem_assoc sym !Fixity.symtbl 
	then let (a_prec,dir) = List.assoc sym !Fixity.symtbl in
	paren a_prec (pp_assoc a_prec dir a term)
	else paren app_prec (pp_atomic (a,[t1;t2]))
    | Atomic a -> paren app_prec (pp_atomic a)
    | Pair (b,c) -> text "(" <:> pp_pair (Pair(b,c)) <:> text ")"
    | Nil -> text "[]" (* Can't tell whether it's the empty string ... *)
    | Cons (CharC _,_) -> 
        (* Try to print as a string constant if it's ground *)
	(match pp_literal term with
	  Some doc -> quotes (doc)
	| None -> bracket (pp_list term))
    | Cons (b,c) -> bracket (pp_list term)
    | Abs (a,c) -> 
	paren abs_prec (Var.pp_var (a) <:> backslash <:> pp abs_prec c)
    | Fresh (t1,t2) -> 
	paren fresh_prec (pp fresh_prec t1 <+> hash <+> pp fresh_prec t2)
    | Eq(t1,t2) -> 
	paren eq_prec (pp eq_prec t1 <+> eq <+> pp eq_prec t2)
    | Is(t1,t2) -> 
	paren is_prec (pp is_prec t1 <+> text "is" <+> pp is_prec t2)
    | Cut -> cut
    | True -> text "true"
    | Implies(t1,t2) -> 
	paren implies_prec 
	  (pp implies_prec  t1 <+> darrow <+> pp implies_prec t2)
    | Or(t1,t2) -> 
	paren or_prec (pp or_prec t1 <+> semi <+> pp or_prec t2)
    | And(t1,t2) -> 
	paren and_prec(pp and_prec t1 <:> comma <+> pp and_prec t2)
    | Forall(v,s,t) -> 
	paren forall_prec 
	  (text "forall" <+> Var.pp_var v <:> colon <:> pp_ty s <:> dot 
	     <+> pp forall_prec  t)
    | Exists(v,s,t) -> 
	paren exists_prec 
	  (text "exists" <+> Var.pp_var v <:> colon <:> pp_ty s <:> dot 
	     <+> pp exists_prec t)
    | New(a,s,t) -> 
	paren new_prec 
	  (text "new" <+> Var.pp_var a <:> colon <:> pp_ty s <:> dot 
	     <+> pp new_prec t)
    | Guard(t1,t2,t3) -> 
	paren guard_prec 
	  (pp guard_prec t1 <+> arrow <+> pp guard_prec t2  
	     <+>  bar <+> pp guard_prec t3)
    | Not(t) -> paren not_prec (text "not" <+> pp not_prec t)
    | Constr(t,c) -> 
	pp constr_prec t  <+> slash <+> pp_constr c
    | Underscore -> underscore
    | IntC i -> num i
    | BoolC c -> if c then text "tt" else text "ff"
    | CharC c -> squotes(escaped_char c)
  and pp_literal t = 
    match t with
      Nil -> Some emp
    | Cons (CharC c, tl) -> 
	(match pp_literal tl with 
	  Some doc -> Some (escaped_char c <:> doc)
	| None -> None)
    | _ -> None
  and pp_pair t = 
    match t with
      Pair(b,c) -> 
	pp pair_prec b <:> text "," <:> pp_pair c
    | t -> pp pair_prec t
  and pp_list t = 
    match t with 
      Nil -> emp
    | Cons (b,Nil) -> 
	pp pair_prec b
    | Cons (b,(Cons (_,_) as c)) ->
	pp pair_prec b <:> text "," <:> pp_list c
    | Cons (b,t) -> pp pair_prec b <:> bar <:> pp pair_prec t
    | t ->  pp pair_prec t
  and pp_assoc prec dir a term = 
    match dir,term with 
      Left,Atomic(b,[t1;t2]) -> 
	if a = b then pp_assoc prec dir a t1 
	    <+> Nstbl.pp_path a 
	    <+> pp prec t2
	else pp prec term
    | Right,Atomic(b,[t1;t2]) -> 
	if a = b then pp prec t1 
	    <+> Nstbl.pp_path a
	    <+> pp_assoc prec dir a t2
	else pp prec term
    | Non,_| _,_  -> pp prec term
  in 
  pp toplevel_prec term
;;

let pp_ty0 ty = 
  match ty with
    UnderscoreTy | IdTy _ | VarTy _ | ListTy _ | UnitTy
  | NameTy _ | IntTy | CharTy | BoolTy | PropTy -> pp_ty ty
  | _ -> paren (pp_ty ty)
;;


let rec pp_subst s = 
  sep newline (List.map 
    (fun (v, e) -> 
      text (Var.to_string' v) 
	<+> arrow 
	<+> pp_term e 
	<:> newline)
    (Varmap.to_list s))
;;

let tab = 8;;

let rec pp_rdecl decl = 
  match decl with
    KindDecl(s,k) -> text s <+> colon <+> pp_kind k <:> dot
  | SymDecl (sym,st) -> text sym <+> colon <+> pp_ty st <:> dot
  | FuncDecl (s,[],ty) ->
    text "cnst" <+> text s <+> eq <+> pp_ty ty <:> dot 
  | FuncDecl (s,tys,ty) ->
    text "func" <+> text s 
	<+> (sep space (List.map pp_ty0 tys)) 
	<+> eq <+> pp_ty ty <:> dot 
  | PredDecl (s,tys) ->
    text "pred" <+> text s <+> 
    (sep space (List.map pp_ty0 tys))<:> dot 
  | ClauseDecl t  -> pp_term t  <:> dot
  | Query (t) -> 
      pp_term t
  | UseDirective(f) -> 
      text "#use" <+> quotes(text f) <:> dot
  | TraceDirective(i) -> 
      text "#trace" <+> quotes(num i) <:> dot 
  | InfixDecl(Left,s,i) -> 
      text "infixl" <+> text s <+> num i <:> dot 
  | InfixDecl(Right,s,i) -> 
      text "infixr" <+> text s <+> num i <:> dot 
  | InfixDecl(Non,s,i) -> 
      text "infixn" <+> text s <+> num i <:> dot 
  | TypeDefn(s,vs,ty) -> 
      text "type" <+> text s <+> (sep space (List.map Var.pp_var vs)) 
	<+> eq <+> pp_ty ty <:> dot 
  | NamespaceDecl (sym,ds) -> 
      text "namespace" <+> text sym 
	<+> paren (indent 8 (sep newline (List.map pp_decl ds))) <:> dot
  | BuiltinFuncDecl (s,[],ty,f) ->
    text "cnst" <+> text s <+> eq <+> pp_ty ty <:> dot 
  | BuiltinFuncDecl (sym,tys,ty,f) -> 
      text "func" <+> text sym 
	<+> (sep space (List.map pp_ty0 tys)) 
	<+> eq <+> pp_ty ty <:> dot 
  | TypeQuery(term) -> 
      text "#type" <+> pp_term term <:> dot
  | QuitDirective -> text "#quit"
  | OpenDirective s -> text "#open" <+> Nstbl.pp_path s
  | HelpDirective(None) -> text "#help"
  | HelpDirective(Some p) -> text "#help"<+> Nstbl.pp_path p
and pp_decl decl = pp_rdecl decl.rdecl
;;


let rec ftvs s = 
  let (+++) = Varset.union in 
  let e = Varset.empty in
  match s with 
    IdTy _ -> Util.impos "ftvs expects ids to have been resolved"
  | DataTy (_,ss) -> Varset.unions (List.map ftvs ss)
  | PairTy(s1,s2) | AbsTy (s1,s2) | ArrowTy(s1,s2) 
    -> ftvs s1 +++ ftvs s2
  | ListTy s -> ftvs s
  | VarTy v -> Varset.singleton v
  | UnitTy | NameTy _ | UnderscoreTy | IntTy 
  | CharTy | PropTy | BoolTy 
    -> e
;;


let k2s = Printer.print_to_string pp_kind;;


let ty2s = Printer.print_to_string pp_ty;;


let t2s = Printer.print_to_string pp_term;;


let rec apply_tysub sb ty = 
  let h = apply_tysub sb in
  match ty with
    NameTy s -> NameTy s
  | DataTy (ds,ss) -> DataTy (ds,List.map h ss)
  | VarTy v -> (try (Varmap.find v sb) with Not_found -> ty)
  | ListTy s -> ListTy(h s)
  | AbsTy (a,s) -> AbsTy(h a, h s)
  | PairTy (s1,s2) -> PairTy (h s1, h s2)
  | IdTy s -> IdTy s
  | UnitTy -> UnitTy
  | UnderscoreTy -> UnderscoreTy
  | IntTy -> IntTy
  | BoolTy -> BoolTy
  | CharTy -> CharTy
  | PropTy -> PropTy
  | ArrowTy(s1,s2) -> ArrowTy(h s1, h s2)
;;


let freshen s = 
  let f = ftvs s in
  let m = Varset.fold 
      (fun v m -> Varmap.add v (VarTy (Var.rename v)) m) f Varmap.empty in
  apply_tysub m s
;;


let rec supp t = 
  let (+++) = Varset.union in 
  let e = Varset.empty in
  match t with
    Atomic(sym,ts) -> Varset.unions (List.map supp ts)
  | Name a -> Varset.singleton a
  | Transpose(a,b,ts) -> Varset.from_list [a;b]
  | Abs(a,t) |New(a,_,t) -> Varset.remove a (supp t)
  | Nil|Unit|True|Cut|Underscore|IntC _|Var _ | CharC _ | BoolC _
    -> e
  | Not t |Forall(_,_,t)|Exists(_,_,t) 
    -> supp t
  | Cons(t1,t2)|Pair(t1,t2)|Fresh (t1,t2)|Eq(t1,t2)|Is(t1,t2)|Implies(t1,t2)
  | Or(t1,t2) | And(t1,t2) 
    -> (supp t1) +++ (supp t2)
  | Guard(t1,t2,t3) -> (supp t1) +++ (supp t2) +++ (supp t3)
  | Constr(t,cs) -> 
      (supp t) +++ (Varset.unions 
		      (List.map (fun (a,_) -> Varset.singleton a) cs))
;;
