(* Declarations and code for builtin types, props, functions. *)
open Absyn;;
module I = Isym;;
open Internal;;
open Nstbl;;

(* Abbreviations matching built-in types with their absyn forms. *)
let unit_ty_decl = TypeDefn("unit",[],UnitTy);;
let int_ty_decl = TypeDefn("int",[],IntTy);;
let char_ty_decl = TypeDefn("char", [], CharTy);;
let bool_ty_decl = TypeDefn("bool", [], BoolTy);;
let bool_tt_decl1 = FuncDecl("tt",[],BoolTy);;
let bool_tt_decl2 = ClauseDecl(Eq(Atomic(Rel(Base("tt")),[]), BoolC true));;
let bool_ff_decl1 = FuncDecl("ff",[],BoolTy);;
let bool_ff_decl2 = ClauseDecl(Eq(Atomic(Rel(Base("ff")),[]), BoolC false));;
let string_ty_decl = TypeDefn("string", [], ListTy (CharTy));;
let a = Var.mkvar' "A";;
let vA = VarTy a;;
let list_ty_decl = TypeDefn("list", [a], ListTy(vA));;
let list_nil_decl1 = FuncDecl("nil",[],ListTy (vA));;
let list_nil_decl2 = ClauseDecl(Eq(Atomic(Rel(Base("nil")),[]), Nil));;
let list_cons_decl1 = FuncDecl("cons",[vA;ListTy(vA)], ListTy(vA));;
let list_cons_decl2 = 
  let x = Var.mkvar' "X" in
  let y = Var.mkvar' "Y" in
  ClauseDecl(Eq(Atomic(Rel(Base("cons")),[Var x; Var y]), 
		Cons(Var x, Var y)));;

let binop ty = ArrowTy(ty,ArrowTy(ty,ty));;
let binrel ty = ArrowTy(ty,ArrowTy(ty,PropTy));;

exception DomainError of string;;

let one_arg f h l = 
  match l with 
    [t] -> 
      h (f t)
  | _ -> raise (DomainError "expected one argument")
;;

let two_args f1 f2 h l = 
  match l with 
    [t1;t2] -> 
      h (f1 t1) (f2 t2)
  | _ -> raise (DomainError "expected two arguments")
;;

let int_arg t = 
  match t with
    App(I.Int i,[]) -> i
  | _ -> raise (DomainError "expected integer argument")
;;

let char_arg t = 
  match t with
    App(I.Char i,[]) -> i
  | _ -> raise (DomainError "expected integer argument")
;;

let bool_arg t = 
  match t with
    App(I.Bool i,[]) -> i
  | _ -> raise (DomainError "expected integer argument")
;;

let unit_arg t = 
  match t with
    App(I.Unit,[]) -> ()
  | _ -> raise (DomainError "expected integer argument")
;;


let rec list_arg a_arg t = 
  match t with
    App(I.Nil,[]) -> []
  | App(I.Cons,[t1;t2]) -> (a_arg t1)::(list_arg a_arg t2)
  | _ -> raise (DomainError "expected list argument")
;;

let int_result x = App(I.Int x,[]);;
let char_result x = App(I.Char x,[]);;
let bool_result x = App(I.Bool x,[]);;
let unit_result x = App(I.Unit,[]);;


let list_result a_result x = 
  List.fold_right (fun h t -> App(I.Cons,[a_result h;t])) x (App(I.Nil,[]))
;;


type 'a ty_rec = {arg : I.term_sym term -> 'a; result:'a -> I.term_sym term; ty: ty};;

let int_rec = {arg=int_arg;result=int_result;ty=IntTy};;
let char_rec = {arg=char_arg;result=char_result;ty=CharTy};;
let unit_rec = {arg=unit_arg;result=unit_result;ty=UnitTy};;
let bool_rec = {arg=bool_arg;result=bool_result;ty=BoolTy};;


let list_rec r = 
  { arg=list_arg r.arg;
    result=list_result r.result;
    ty=ListTy r.ty}
;;


let string_rec = list_rec char_rec;;

let ( ** ) f g x = f (g x);;


let do_unop t1 t s f = 
  BuiltinFuncDecl(s,[t1.ty],t.ty,
		  t.result ** (one_arg t1.arg (f)))
;;


let do_binop t1 t2 t s f = 
  BuiltinFuncDecl(s,[t1.ty;t2.ty],t.ty,
		  t.result ** (two_args t1.arg t2.arg (f)))
;;


let do_bool_binop s f = do_binop bool_rec bool_rec bool_rec s f;;
let do_int_binop s f = do_binop int_rec int_rec int_rec s f;;
let do_int_binrel s f = do_binop int_rec int_rec bool_rec s f;;

let plus_decl = do_int_binop "plus" (+);;
let minus_decl = do_int_binop "minus" (-);;
let times_decl = do_int_binop "times" ( * );;
let div_decl = do_int_binop "div" (/);;
let mod_decl = do_int_binop "mod" (mod);;
let min_decl = do_int_binop "min" min;;
let max_decl = do_int_binop "max" max;;
let lt_decl = do_int_binrel "lt" (<);;
let leq_decl = do_int_binrel "leq" (<=);;
let gt_decl = do_int_binrel "gt" (>);;
let geq_decl = do_int_binrel "geq" (>=);;

let and_decl = do_bool_binop "and" (&&);;
let or_decl = do_bool_binop "or" (||);;


let explode s = 
  let n = String.length s in
  let rec expl i = 
    if i < n then 
      String.get s i :: expl (i+1)
    else []
  in expl 0
;;


let implode cs = 
  let n = List.length cs in 
  let s = String.create n in
  let rec impl i cs = 
    if i < n 
    then 
      (String.set s i (List.hd cs); impl (i+1) (List.tl cs))
    else ()
  in impl 0 cs; s
;;



(* Getting rid of this impure kind of side-effecting *)
(*
let print_decl = do_unop string_rec unit_rec "print" (fun l -> print_string(implode(l)));;

let read_ln_decl = do_unop unit_rec string_rec "read_line" (fun () -> explode(read_line()));;

let nl_decl = do_unop unit_rec unit_rec "nl" Util.nl;;
*)

let mk_decl rd = {pos = None; rdecl = rd};;

let decls = List.map mk_decl [NamespaceDecl("base",
					     List.map mk_decl 
					      [ unit_ty_decl;
						int_ty_decl;
						string_ty_decl;
						bool_ty_decl;
						bool_tt_decl1;
						bool_tt_decl2;
						bool_ff_decl1;
						bool_ff_decl2;
						char_ty_decl;
						list_ty_decl;
						list_nil_decl1;
						list_nil_decl2;
						list_cons_decl1;
						list_cons_decl2;
						plus_decl;
						minus_decl;
						times_decl;
						div_decl;
						min_decl;
						max_decl;
						lt_decl;
						leq_decl;
						gt_decl;
						geq_decl;
(*						print_decl;
						read_ln_decl;
						nl_decl;*)
						and_decl;
						or_decl]
					      );
			       OpenDirective(Rel(Base("base")));
			     ];;


let bindings = Hashtbl.create 100;;
open Printer;;

let bind u f = Hashtbl.add bindings u f;;
let apply c ts = 
  match c with
    I.Symb (u,_) -> 
      if Hashtbl.mem bindings u 
      then (Hashtbl.find bindings u) ts
      else App(c,ts)
  | c -> App(c,ts)
;;
