(* Alpha Prolog *)
open Types;;
open Internal;;
module P = Perm;;
module S = Subst;;
open Printer;;
    
(* TODO: Use a faster version of unification such as using reference cells *)
    
(* TODO: Implement "complete" unification for arbitrary terms. *)
    
(* TODO: Interleave freshness constraint solving with unification whenever a 
    substitution is performed (solve constraints "eagerly" *)
    
type 'a constrs = ('a term * var) list;;


type 'a fprobs = ('a term * 'a term) list;;


type 'a answer = ('a S.subst * 'a constrs);;


(* return value indicates whether failure was due to mismatched atoms or
   some other reason: this is used to determine whether to try a
   different atom instantiation *)
type 'a sc_transform = ('a -> bool) -> ('a -> bool);;


type 'a equal = 'a -> 'a -> bool;;


let rec compress t sub = 
  match t with
    Susp(p,Exist b,x) -> 
      (match S.lookup sub x with 
	  None -> t
	| Some t' -> compress (apply_p p t') sub)
  | t -> t
;;


let rec unify_fresh1 t1 t2 sub sc constrs = 
  match compress t1 sub, compress t2 sub with
  | Name a,Name b -> 
      if Var.eq a b 
      then true
      else sc constrs
  | Name a,Abs(b,e) -> 
      if Var.eq a b 
      then sc constrs
      else 
	unify_fresh1 t1 e sub sc constrs
(*  | Susp(p,b,v), Abs (a,t) -> 
      (* try the "good" (non-substitution inducing) case: A # t. *)
      let _ = unify_fresh1 t1 t sub sc constrs in
      Util.impos "Unhandled variable freshness case: var # abs" *)
  | _,App(c,es) -> 
      unifys_fresh (List.map (fun t2 -> (t1,t2)) es) sub sc constrs
  | Susp(p,b,x), Name a 
  | Name a, Susp(p,b,x) ->  
      let a' = (P.lookup (P.inv p) a) in
      sc ((Name a', x)::constrs)
  | Susp(p,b,x), Susp(q,c,y) -> 
      (sc ((Susp(Perm.comp (Perm.inv q) p, b,x), y)::constrs))
  | _,_ -> Util.impos "Internal error: illegal case in freshness unification"

and unifys_fresh fs sub sc constrs = 
  match fs with
    [] -> sc constrs 
  | (a,e)::fs -> unify_fresh1 a e sub (unifys_fresh fs sub sc) constrs
;;


let do_susp p1 p2 x sc (sub,fs) = 
  let ds = P.disagreement_set p1 p2 in
  let new_fs = List.map (fun a -> (Name a,x)) ds in
  sc (sub, new_fs@fs)
;;


let bconstr_leq c1 c2 = 
  match c1,c2 with
    _,Toplevel -> true
  | Constrained l, Constrained l' -> List.length l <= List.length l'
  | _,_ -> false
;;


let bconstr_satisfies x c = 
  match c with
    Toplevel -> true
  | Constrained l -> List.exists (Var.eq x) l
;;


(* Extended occurs check: check that x does not occur in t, that
   all Uvars in t are permitted for x, and constrain all Evars in 
   t to be as constrained as x is.  *)
let rec occurs_check x c t sc subst = 
  match compress t subst with
    Name a -> 
      if !Flags.strict_new_occurs
      then 
        if bconstr_satisfies a c
        then sc (t,subst)
        else true (* maybe this should be false ? *)
      else sc (t,subst)
  | Abs(a,e) -> 
      occurs_check x c e (fun (t',subst) -> 
      sc (Abs(a,t'), subst)) subst
  | App(f,es) -> 
      occurs_checks x c es (fun (ts,subst) -> 
      sc (App(f,es),subst)) subst
  | Susp(p,Univ,y) -> 
      if bconstr_satisfies y c
      then sc (t,subst)
      else true (* might be because of a name *)
  | Susp(p,Exist c',y) -> 
      if Var.eq x y 
      then false
      else 
        if bconstr_leq c' c (* If lhs more constrained than rhs *)
	then sc (t,subst) (* continue *)
	else (sc (Susp(p,Exist c, y),subst)) 
and occurs_checks x c ts sc subst = 
  match ts with
    [] -> sc ([],subst)
  | t::ts -> 
      occurs_check x c t (fun (t',subst) -> 
      occurs_checks x c ts (fun (ts',subst) -> 
      sc (t'::ts',subst)) subst) subst
;;
	
let unify_wrap eq t1 t2 sc ans = 
  let rec do_subst t p b x sc (subst,constrs) = 
    match S.lookup subst x with
      None -> 
	occurs_check x b t (fun (t',subst) -> 
	let subst' = S.join x (apply_p (P.inv p) t') subst in
	let new_fs = List.map (fun (t,u) -> 
	(S.apply_s subst' t, 
	 S.apply_s subst' (Internal.Susp(Perm.id,Exist(Toplevel),u)))) 
	    constrs in
	unifys_fresh new_fs subst' (fun constrs' -> 
	  sc (subst',constrs')) []
	  ) subst
    | Some t' -> 
	unify1 t (apply_p p t') sc (subst, constrs) 
	  
  and do_EE_subst (p1,c1,x) (p2,c2,y) sc ans = 
    do_subst (Susp(p2,Exist c2,y)) p1 c1 x sc ans 
      
  and unify1 t1 t2 sc ans = 
    match compress t1 (fst ans), compress t2 (fst ans) with
      (Name a, Name b) -> 
	if Var.eq a b 
	then sc ans
	else true
    | (Abs(a,e), Abs(b,e')) -> 
	if Var.eq a b 
	then unify1 e e' sc ans
	else 
	  unify1 e (apply_p (P.trans a b) e') 
	    (fun (s,c) -> unify_fresh1 (Name a) e' s (fun c' -> sc (s,c')) c) ans
    | (App(c,es),App(d,es')) -> 
	if eq c d 
	then unifys es es' sc ans
	else false
    | (Susp(p1,Univ,x),Susp(p2,Univ,y)) -> 
	if Var.eq x y && P.disagreement_set p1 p2 = []
	then sc ans 
	else false
    | (Susp(p1,Exist b1,x),Susp(p2,Exist b2,y)) -> 
	if Var.eq x y 
	then do_susp p1 p2 x sc ans
	else do_EE_subst (p1,b1,x) (p2,b2,y) sc ans
    | (Susp(p,Exist b,x), t) -> 
	do_subst t p b x sc ans
    | (t,Susp(p,Exist b,x)) -> 
	do_subst t p b x sc ans
    | _,_ -> false
  and unifys es es' sc ans = 
    match es, es' with
      [],[] -> sc ans
    | e::es,e'::es' -> 
	unify1 e e' (fun ans -> 
	  unifys es es' sc ans) ans
    | _,_ -> false
	  
  in 
  unify1 t1 t2 sc ans
;;


let unify_fresh a t sc (sub,constrs) = 
  unify_fresh1 a t sub (fun constrs' -> sc (sub,constrs')) constrs
;;


let unify eq e1 e2 sc (sub,constrs) = 
  unify_wrap eq e1 e2 sc (sub,constrs)
(*  unify1 eq e1 e2 
    (fun (sub,fs) -> 
      unifys_fresh fs sub 
	(fun constrs' -> sc (sub,constrs'))
	constrs) (sub,[]) *)
;;


let rec unifys eq es sc = 
  match es with
    [] -> sc
  | (t1,t2)::es -> unify eq t1 t2 (unifys eq es sc)
;;


let rec pp_constrs pp constrs = 
  match constrs with
    [] -> emp
  | (t,x)::constrs -> 
      pp_term pp t <+>
      text "#" <+>
      text (Var.to_string' x) <:>
      newline <:>
      pp_constrs pp constrs
;;

	


let rec pp_answer sg env fvs (subst,constrs) = 
  match fvs with
   [] -> 
     pp_constrs Isym.pp_term_sym constrs 
   | x::xs -> 
     match S.finish subst x with
       None -> pp_answer sg env xs (subst,constrs) 
     | Some t -> 
         let t' = Translate.untranslate_tm sg env t in 
         text (Var.to_string' x) <+> 
         text "=" <+>
         Absyn.pp_term t' <+>
         newline <:>
         pp_answer sg env xs (subst, constrs)
;;


